@@ -75,6 +75,7 @@ enum {
         PROP_HOST_IP,
         PROP_PORT,
         PROP_SERVER,
+        PROP_SESSION,
         PROP_SUBSCRIPTION_TIMEOUT
 };
 
@@ -183,13 +184,50 @@ get_default_host_ip (void)
         unsigned long dest;
         gboolean found = FALSE;
         
+#if defined(__FreeBSD__)
+	if ((fp = popen ("netstat -r -f inet -n -W", "r"))) {
+		char buffer[BUFSIZ];
+
+		char destination[32];
+
+		int i;
+		/* Skip the 4 header lines */
+		for (i=0;i<4;i++) {
+			if (!(fgets(buffer, BUFSIZ, fp)))
+				return NULL; /* Can't read */
+
+			if (buffer[strlen(buffer)-1] != '\n') {
+				g_warning("Can't read netstat output!");
+				return NULL;
+			}
+		}
+
+		while (fgets(buffer, BUFSIZ, fp)) {
+			if (buffer[strlen(buffer)-1] != '\n') {
+				g_warning("Can't read netstat output!");
+				return NULL;
+			}
+
+			if (sscanf(buffer, "%s %*s %*s %*d %*d %*d %s %*d", destination, dev) == 2) {
+				if (strcmp("default", destination) == 0) {
+					found = TRUE;
+					break;
+				}
+			}
+		}
+		pclose(fp);
+	}
+#else
         /* TODO: error checking */
 
         fp = fopen ("/proc/net/route", "r");
         
         /* Skip the header */
-        fscanf (fp, "%*[^\n]\n");
-        
+        if (fscanf (fp, "%*[^\n]\n") == EOF) {
+               fclose (fp);
+               return NULL;
+	}
+
         while ((ret = fscanf (fp,
                               "%31s %lx %*x %*X %*d %*d %*d %*x %*d %*d %*d",
                               dev, &dest)) != EOF) {
@@ -202,6 +240,7 @@ get_default_host_ip (void)
         }
         
         fclose (fp);
+#endif
 
         return get_host_ip (found ? dev : NULL);
 }
@@ -216,22 +255,37 @@ gupnp_context_init (GUPnPContext *context)
                                              GUPNP_TYPE_CONTEXT,
                                              GUPnPContextPrivate);
 
-        context->priv->session = soup_session_async_new ();
-
-        g_object_set (context->priv->session,
-                      SOUP_SESSION_IDLE_TIMEOUT,
-                      60,
-                      NULL);
-
         server_id = make_server_id ();
         gssdp_client_set_server_id (GSSDP_CLIENT (context), server_id);
         g_free (server_id);
+}
+
+static GObject *
+gupnp_context_constructor (GType                  type,
+                           guint                  n_props,
+                           GObjectConstructParam *props)
+{
+	GObject *object;
+	GUPnPContext *context;
+
+	object = G_OBJECT_CLASS (gupnp_context_parent_class)->constructor
+                (type, n_props, props);
+	context = GUPNP_CONTEXT (object);
+
+        context->priv->session = soup_session_async_new_with_options
+                (SOUP_SESSION_IDLE_TIMEOUT,
+                 60,
+                 SOUP_SESSION_ASYNC_CONTEXT,
+                 gssdp_client_get_main_context (GSSDP_CLIENT (context)),
+                 NULL);
 
 	if (g_getenv ("GUPNP_DEBUG")) {
 		SoupLogger *logger;
 		logger = soup_logger_new (SOUP_LOGGER_LOG_BODY, -1);
 		soup_logger_attach (logger, context->priv->session);
 	}
+
+	return object;
 }
 
 static void
@@ -283,6 +337,10 @@ gupnp_context_get_property (GObject    *object,
                 g_value_set_object (value,
                                     gupnp_context_get_server (context));
                 break;
+        case PROP_SESSION:
+                g_value_set_object (value,
+                                    gupnp_context_get_session (context));
+                break;
         case PROP_SUBSCRIPTION_TIMEOUT:
                 g_value_set_uint (value,
                                   gupnp_context_get_subscription_timeout
@@ -341,6 +399,7 @@ gupnp_context_class_init (GUPnPContextClass *klass)
 
         object_class = G_OBJECT_CLASS (klass);
 
+        object_class->constructor  = gupnp_context_constructor;
         object_class->set_property = gupnp_context_set_property;
         object_class->get_property = gupnp_context_get_property;
         object_class->dispose      = gupnp_context_dispose;
@@ -401,6 +460,23 @@ gupnp_context_class_init (GUPnPContextClass *klass)
                                       G_PARAM_STATIC_NICK |
                                       G_PARAM_STATIC_BLURB));
 
+        /**
+         * GUPnPContext:session
+         *
+         * The #SoupSession object used by GUPnP.
+         **/
+        g_object_class_install_property
+                (object_class,
+                 PROP_SESSION,
+                 g_param_spec_object ("session",
+                                      "SoupSession",
+                                      "SoupSession object",
+                                      SOUP_TYPE_SESSION,
+                                      G_PARAM_READABLE |
+                                      G_PARAM_STATIC_NAME |
+                                      G_PARAM_STATIC_NICK |
+                                      G_PARAM_STATIC_BLURB));
+
         /**
          * GUPnPContext:subscription-timeout
          *
@@ -424,9 +500,20 @@ gupnp_context_class_init (GUPnPContextClass *klass)
                                     G_PARAM_STATIC_BLURB));
 }
 
+/**
+ * gupnp_context_get_session
+ * @context: A #GUPnPContext
+ *
+ * Get the #SoupSession object that GUPnP is using.
+ *
+ * Return value: The #SoupSession used by GUPnP. Do not unref this when
+ * finished.
+ **/
 SoupSession *
-_gupnp_context_get_session (GUPnPContext *context)
+gupnp_context_get_session (GUPnPContext *context)
 {
+        g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
+
         return context->priv->session;
 }
 
@@ -458,9 +545,12 @@ gupnp_context_get_server (GUPnPContext *context)
         g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
 
         if (context->priv->server == NULL) {
-                context->priv->server = soup_server_new (SOUP_SERVER_PORT,
-                                                         context->priv->port,
-                                                         NULL);
+                context->priv->server = soup_server_new
+                        (SOUP_SERVER_PORT,
+                         context->priv->port,
+                         SOUP_SERVER_ASYNC_CONTEXT,
+                         gssdp_client_get_main_context (GSSDP_CLIENT (context)),
+                         NULL);
 
                 soup_server_add_handler (context->priv->server, NULL,
                                          default_server_handler, context,
@@ -678,10 +768,11 @@ host_path_handler (SoupServer        *server,
         char *local_path, *path_to_open;
         struct stat st;
         int status;
-        GList *locales;
+        GList *locales, *orig_locales;
         GMappedFile *mapped_file;
         GError *error;
 
+        orig_locales = NULL;
         locales      = NULL;
         local_path   = NULL;
         path_to_open = NULL;
@@ -702,7 +793,7 @@ host_path_handler (SoupServer        *server,
         }
 
         /* Get preferred locales */
-        locales = http_request_get_accept_locales (msg);
+        orig_locales = locales = http_request_get_accept_locales (msg);
 
  AGAIN:
         /* Add locale suffix if available */
@@ -860,9 +951,9 @@ host_path_handler (SoupServer        *server,
         g_free (path_to_open);
         g_free (local_path);
 
-        while (locales) {
-                g_free (locales->data);
-                locales = g_list_delete_link (locales, locales);
+        while (orig_locales) {
+                g_free (orig_locales->data);
+                orig_locales = g_list_delete_link (orig_locales, orig_locales);
         }
 }
 
