@@ -90,7 +90,7 @@ typedef struct {
 
         int           seq;
 
-        guint         timeout_id;
+        GSource      *timeout_src;
 
         GList        *pending_messages; /* Pending SoupMessages from this
                                            subscription */
@@ -104,7 +104,7 @@ subscription_data_free (SubscriptionData *data)
 
         context = gupnp_service_info_get_context
                         (GUPNP_SERVICE_INFO (data->service));
-        session = _gupnp_context_get_session (context);
+        session = gupnp_context_get_session (context);
 
         /* Cancel pending messages */
         while (data->pending_messages) {
@@ -130,8 +130,8 @@ subscription_data_free (SubscriptionData *data)
 
         g_free (data->sid);
 
-        if (data->timeout_id)
-                g_source_remove (data->timeout_id);
+        if (data->timeout_src)
+	        g_source_destroy (data->timeout_src);
 
         g_slice_free (SubscriptionData, data);
 }
@@ -151,13 +151,16 @@ notify_data_free (NotifyData *data)
 }
 
 struct _GUPnPServiceAction {
-        const char  *name;
+        GUPnPContext *context;
 
-        SoupMessage *msg;
+        char         *name;
 
-        xmlNode     *node;
+        SoupMessage  *msg;
 
-        GString     *response_str;
+        xmlDoc       *doc;
+        xmlNode      *node;
+
+        GString      *response_str;
 };
 
 GType
@@ -172,6 +175,59 @@ gupnp_service_action_get_type (void)
         return our_type;
 }
 
+static void
+finalize_action (GUPnPServiceAction *action)
+{
+        SoupServer *server;
+        char *response_body;
+
+        /* Embed action->response_str in a SOAP document */
+        g_string_prepend (action->response_str,
+                          "<?xml version=\"1.0\"?>"
+                          "<s:Envelope xmlns:s="
+                                "\"http://schemas.xmlsoap.org/soap/envelope/\" "
+                          "s:encodingStyle="
+                                "\"http://schemas.xmlsoap.org/soap/encoding/\">"
+                          "<s:Body>");
+
+        if (action->msg->status_code != SOUP_STATUS_INTERNAL_SERVER_ERROR) {
+                g_string_append (action->response_str, "</u:");
+                g_string_append (action->response_str, action->name);
+                g_string_append (action->response_str, "Response>");
+        }
+
+        g_string_append (action->response_str,
+                         "</s:Body>"
+                         "</s:Envelope>");
+
+        response_body = g_string_free (action->response_str, FALSE);
+
+        soup_message_set_response (action->msg,
+                                   "text/xml",
+                                   SOUP_MEMORY_TAKE,
+                                   response_body,
+                                   strlen (response_body));
+
+        /* Server header on response */
+        soup_message_headers_append
+                        (action->msg->response_headers,
+                         "Server",
+                         gssdp_client_get_server_id
+                                (GSSDP_CLIENT (action->context)));
+
+        /* Tell soup server that response is now ready */
+        server = gupnp_context_get_server (action->context);
+        soup_server_unpause_message (server, action->msg);
+
+        /* Cleanup */
+        g_free (action->name);
+        g_object_unref (action->msg);
+        g_object_unref (action->context);
+        xmlFreeDoc (action->doc);
+
+        g_slice_free (GUPnPServiceAction, action);
+}
+
 /**
  * gupnp_service_action_get_name
  * @action: A #GUPnPServiceAction
@@ -415,6 +471,8 @@ gupnp_service_action_return (GUPnPServiceAction *action)
         g_return_if_fail (action != NULL);
 
         soup_message_set_status (action->msg, SOUP_STATUS_OK);
+
+        finalize_action (action);
 }
 
 /**
@@ -496,6 +554,8 @@ gupnp_service_action_return_error (GUPnPServiceAction *action,
 
         soup_message_set_status (action->msg,
                                  SOUP_STATUS_INTERNAL_SERVER_ERROR);
+
+        finalize_action (action);
 }
 
 static void
@@ -602,7 +662,7 @@ query_state_variable (GUPnPService       *service,
 
 /* controlURL handler */
 static void
-control_server_handler (SoupServer        *soup_server,
+control_server_handler (SoupServer        *server,
                         SoupMessage       *msg, 
                         const char        *server_path,
                         GHashTable        *query,
@@ -611,12 +671,10 @@ control_server_handler (SoupServer        *soup_server,
 {
         GUPnPService *service;
         GUPnPContext *context;
-        GSSDPClient *client;
         xmlDoc *doc;
         xmlNode *action_node;
         const char *soap_action, *action_name;
         char *end;
-        gpointer response_body;
         GUPnPServiceAction *action;
 
         service = GUPNP_SERVICE (user_data);
@@ -628,7 +686,6 @@ control_server_handler (SoupServer        *soup_server,
         }
 
         context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (service));
-        client = GSSDP_CLIENT (context);
 
         /* Get action name */
         soap_action = soup_message_headers_get (msg->request_headers,
@@ -647,8 +704,8 @@ control_server_handler (SoupServer        *soup_server,
         *end = '\0';
 
         /* Parse action_node */
-        doc = xmlParseMemory (msg->request_body->data,
-                              msg->request_body->length);
+        doc = xmlRecoverMemory (msg->request_body->data,
+                                msg->request_body->length);
         action_node = xml_util_get_element ((xmlNode *) doc,
                                             "Envelope",
                                             "Body",
@@ -663,64 +720,40 @@ control_server_handler (SoupServer        *soup_server,
         /* Create action structure */
         action = g_slice_new (GUPnPServiceAction);
 
-        action->name         = action_name;
-        action->msg          = msg;
+        action->name         = g_strdup (action_name);
+        action->msg          = g_object_ref (msg);
+        action->doc          = doc;
         action->node         = action_node;
         action->response_str = new_action_response_str (service, action_name);
+        action->context      = g_object_ref (context);
+
+        /* Tell soup server that response is not ready yet */
+        soup_server_pause_message (server, msg);
 
         /* QueryStateVariable? */
         if (strcmp (action_name, "QueryStateVariable") == 0)
                 query_state_variable (service, action);
         else {
-                /* Emit signal. Handler parses request and fills in response. */
-                g_signal_emit (service,
-                               signals[ACTION_INVOKED],
-                               g_quark_from_string (action_name),
-                               action);
+                GQuark action_name_quark;
 
-                /* Was it handled? */
-                if (msg->status_code == SOUP_STATUS_NONE) {
-                        /* No. */
+                action_name_quark = g_quark_from_string (action_name);
+                if (g_signal_has_handler_pending (service,
+                                                  signals[ACTION_INVOKED],
+                                                  action_name_quark,
+                                                  FALSE)) {
+                        /* Emit signal. Handler parses request and fills in
+                         * response. */
+                        g_signal_emit (service,
+                                       signals[ACTION_INVOKED],
+                                       action_name_quark,
+                                       action);
+                } else {
+                        /* No handlers attached. */
                         gupnp_service_action_return_error (action,
                                                            401,
                                                            "Invalid Action");
                 }
         }
-
-        /* Embed action->response_str in a SOAP document */
-        g_string_prepend (action->response_str, 
-                          "<?xml version=\"1.0\"?>"
-                          "<s:Envelope xmlns:s="
-                                "\"http://schemas.xmlsoap.org/soap/envelope/\" "
-                          "s:encodingStyle="
-                                "\"http://schemas.xmlsoap.org/soap/encoding/\">"
-                          "<s:Body>");
-
-        if (action->msg->status_code != SOUP_STATUS_INTERNAL_SERVER_ERROR) {
-                g_string_append (action->response_str, "</u:");
-                g_string_append (action->response_str, action_name);
-                g_string_append (action->response_str, "Response>");
-        }
-
-        g_string_append (action->response_str,
-                         "</s:Body>"
-                         "</s:Envelope>");
-
-        response_body = g_string_free (action->response_str, FALSE);
-
-        soup_message_set_response (msg,
-                                   "text/xml",
-                                   SOUP_MEMORY_TAKE,
-                                   response_body,
-                                   strlen (response_body));
-
-        /* Cleanup */
-        g_slice_free (GUPnPServiceAction, action);
-
-        /* Server header on response */
-        soup_message_headers_append (msg->response_headers,
-                                     "Server",
-                                     gssdp_client_get_server_id (client));
 }
 
 /* Generates a standard (re)subscription response */
@@ -858,9 +891,23 @@ subscribe (GUPnPService *service,
         /* Add timeout */
         timeout_seconds = parse_and_limit_timeout (timeout);
         if (timeout_seconds > 0) {
-                data->timeout_id = g_timeout_add_seconds (timeout_seconds,
-                                                          subscription_timeout,
-                                                          data);
+                GUPnPContext *context;
+                GMainContext *main_context;
+
+	        data->timeout_src =
+                        g_timeout_source_new_seconds (timeout_seconds);
+	        g_source_set_callback (data->timeout_src,
+                                       subscription_timeout,
+                                       data,
+                                       NULL);
+
+                context = gupnp_service_info_get_context
+                        (GUPNP_SERVICE_INFO (service));
+                main_context = gssdp_client_get_main_context
+                        (GSSDP_CLIENT (context));
+	        g_source_attach (data->timeout_src, main_context);
+
+                g_source_unref (data->timeout_src);
         }
 
         /* Add to hash */
@@ -923,17 +970,30 @@ resubscribe (GUPnPService *service,
         }
 
         /* Update timeout */
-        if (data->timeout_id) {
-                g_source_remove (data->timeout_id);
-                data->timeout_id = 0;
+        if (data->timeout_src) {
+                g_source_destroy (data->timeout_src);
+                data->timeout_src = NULL;
         }
 
         timeout_seconds = parse_and_limit_timeout (timeout);
         if (timeout_seconds > 0) {
-                data->timeout_id = g_timeout_add_seconds (timeout_seconds,
-                                                          subscription_timeout,
-                                                          data);
-        }
+                GUPnPContext *context;
+                GMainContext *main_context;
+
+	        data->timeout_src = g_timeout_source_new_seconds
+                        (timeout_seconds);
+	        g_source_set_callback(data->timeout_src,
+				      subscription_timeout,
+				      data, NULL);
+
+                context = gupnp_service_info_get_context
+                        (GUPNP_SERVICE_INFO (service));
+                main_context = gssdp_client_get_main_context
+                        (GSSDP_CLIENT (context));
+	        g_source_attach (data->timeout_src, main_context);
+
+                g_source_unref (data->timeout_src);
+	}
 
         /* Respond */
         subscription_response (service, msg, sid, timeout_seconds);
@@ -953,7 +1013,7 @@ unsubscribe (GUPnPService *service,
 
 /* eventSubscriptionURL handler */
 static void
-subscription_server_handler (SoupServer        *soup_server,
+subscription_server_handler (SoupServer        *server,
                              SoupMessage       *msg, 
                              const char        *server_path,
                              GHashTable        *query,
@@ -1152,8 +1212,10 @@ gupnp_service_set_property (GObject      *object,
 
         switch (property_id) {
         case PROP_ROOT_DEVICE:
-                service->priv->root_device =
-                        g_object_ref (g_value_get_object (value));
+                service->priv->root_device = g_value_get_object (value);
+                g_object_add_weak_pointer
+                        (G_OBJECT (service->priv->root_device),
+                         (gpointer *) &service->priv->root_device);
 
                 service->priv->notify_available_id =
                         g_signal_connect_object (service->priv->root_device,
@@ -1207,7 +1269,10 @@ gupnp_service_dispose (GObject *object)
                                  service->priv->notify_available_id);
                 }
 
-                g_object_unref (service->priv->root_device);
+                g_object_remove_weak_pointer
+                        (G_OBJECT (service->priv->root_device),
+                         (gpointer *) &service->priv->root_device);
+
                 service->priv->root_device = NULL;
         }
 
@@ -1292,7 +1357,8 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
          * @action: The invoked #GUPnPAction
          *
          * Emitted whenever an action is invoked. Handler should process
-         * @action.
+         * @action and must call either gupnp_service_action_return() or
+         * gupnp_service_action_return_error().
          **/
         signals[ACTION_INVOKED] =
                 g_signal_new ("action-invoked",
@@ -1466,16 +1532,17 @@ notify_got_response (SoupSession *session,
 
                         context = gupnp_service_info_get_context
                                         (GUPNP_SERVICE_INFO (data->service));
-                        session = _gupnp_context_get_session (context);
+                        session = gupnp_context_get_session (context);
 
                         soup_session_requeue_message (session, msg);
                 } else {
                         /* Emit 'notify-failed' signal */
                         GError *error;
 
-                        error = g_error_new (GUPNP_EVENTING_ERROR,
-                                             GUPNP_EVENTING_ERROR_NOTIFY_FAILED,
-                                             msg->reason_phrase);
+                        error = g_error_new_literal
+                                        (GUPNP_EVENTING_ERROR,
+                                         GUPNP_EVENTING_ERROR_NOTIFY_FAILED,
+                                         msg->reason_phrase);
 
                         g_signal_emit (data->service,
                                        signals[NOTIFY_FAILED],
@@ -1542,7 +1609,7 @@ notify_subscriber (gpointer key,
 
         /* Add body */
         soup_message_set_request (msg,
-                                  "text/xml",
+                                  "text/xml; charset=utf-8",
                                   SOUP_MEMORY_TAKE,
                                   g_strdup (property_set),
                                   strlen (property_set));
@@ -1552,7 +1619,7 @@ notify_subscriber (gpointer key,
 
         context = gupnp_service_info_get_context
                         (GUPNP_SERVICE_INFO (data->service));
-        session = _gupnp_context_get_session (context);
+        session = gupnp_context_get_session (context);
 
         soup_session_queue_message (session,
                                     msg,
